from datetime import datetime, date, time
from decimal import Decimal
from typing import Type, Any, Dict, Optional
import pytest

from compressedfhir.utilities.json_serializers.type_preservation_decoder import (
    TypePreservationDecoder,
)


class TestCustomObject:
    def __init__(
        self,
        name: str,
        value: int,
        created_at: Optional[datetime] = None,
        nested_data: Optional[Dict[str, Any]] = None,
    ):
        self.name: str = name
        self.value: int = value
        self.created_at: Optional[datetime] = created_at
        self.nested_data: Optional[Dict[str, Any]] = nested_data

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TestCustomObject):
            return False
        return (
            self.name == other.name
            and self.value == other.value
            and self.created_at == other.created_at
            and self.nested_data == other.nested_data
        )


@pytest.mark.parametrize(
    "input_type, input_dict, expected_type",
    [
        (
            "datetime",
            {"__type__": "datetime", "iso": "2023-01-01T00:00:00+00:00"},
            datetime,
        ),
        (
            "datetime",
            {
                "__type__": "datetime",
                "iso": "2023-01-01T00:00:00-08:00",
                "tzinfo": "Pacific/Honolulu",
            },
            datetime,
        ),
        ("date", {"__type__": "date", "iso": "2023-01-01"}, date),
        ("time", {"__type__": "time", "iso": "14:30:15"}, time),
        (
            "time",
            {"__type__": "time", "iso": "14:30:15", "tzinfo": "Pacific/Honolulu"},
            time,
        ),
        ("decimal", {"__type__": "decimal", "value": "3.14"}, Decimal),
        ("complex", {"__type__": "complex", "real": 3, "imag": 4}, complex),
        ("bytes", {"__type__": "bytes", "value": "test"}, bytes),
        ("set", {"__type__": "set", "values": [1, 2, 3]}, set),
    ],
)
def test_complex_type_decoding(
    input_type: str, input_dict: Dict[str, Any], expected_type: Type[Any]
) -> None:
    """
    Test decoding of various complex types
    """
    decoded = TypePreservationDecoder.decode(input_dict)
    assert isinstance(decoded, expected_type)


def test_custom_object_decoding() -> None:
    """
    Test decoding of custom objects
    """
    custom_obj_dict = {
        "__type__": "TestCustomObject",
        "__module__": __name__,
        "attributes": {"name": "test", "value": 42},
    }
    decoded = TypePreservationDecoder.decode(custom_obj_dict)
    assert isinstance(decoded, TestCustomObject)
    assert decoded.name == "test"
    assert decoded.value == 42


def test_custom_decoder() -> None:
    """
    Test custom decoder functionality
    """

    def custom_decoder(data: Dict[str, Any]) -> Any:
        if data.get("__type__") == "special_type":
            return f"Decoded: {data['value']}"
        return data

    special_dict = {"__type__": "special_type", "value": "test"}
    decoded = TypePreservationDecoder.decode(
        special_dict, custom_decoders={"special_type": custom_decoder}
    )
    assert decoded == "Decoded: test"


def test_nested_datetime_decoding() -> None:
    """
    Test decoding of nested datetime fields
    """
    nested_datetime_dict = {
        "__type__": "TestCustomObject",
        "__module__": __name__,
        "attributes": {
            "name": "test",
            "value": 42,
            "created_at": {"__type__": "datetime", "iso": "2023-06-15T10:30:00"},
            "nested_data": {
                "timestamp": {"__type__": "datetime", "iso": "2023-06-16T15:45:00"}
            },
        },
    }

    decoded: TestCustomObject = TypePreservationDecoder.decode(nested_datetime_dict)

    assert isinstance(decoded, TestCustomObject)
    assert decoded.name == "test"
    assert decoded.value == 42

    # Check nested datetime fields
    assert hasattr(decoded, "created_at")
    assert isinstance(decoded.created_at, datetime)
    assert decoded.created_at.year == 2023
    assert decoded.created_at.month == 6
    assert decoded.created_at.day == 15

    assert hasattr(decoded, "nested_data")
    assert isinstance(decoded.nested_data, dict)
    assert "timestamp" in decoded.nested_data
    assert isinstance(decoded.nested_data["timestamp"], datetime)
    assert decoded.nested_data["timestamp"].year == 2023
    assert decoded.nested_data["timestamp"].month == 6
    assert decoded.nested_data["timestamp"].day == 16


def test_direct_value_decoding() -> None:
    """
    Test decoding of direct values without type markers
    """
    # Test datetime direct string
    datetime_str = "2023-01-01T00:00:00"
    decoded_datetime = TypePreservationDecoder.decode(datetime_str)
    assert decoded_datetime == datetime_str

    # Test list with mixed types
    mixed_list = [
        {"__type__": "datetime", "iso": "2023-06-15T10:30:00"},
        42,
        "plain string",
    ]
    decoded_list = TypePreservationDecoder.decode(mixed_list)
    assert len(decoded_list) == 3
    assert isinstance(decoded_list[0], datetime)
    assert decoded_list[1] == 42
    assert decoded_list[2] == "plain string"



import logging
from collections import OrderedDict
from datetime import datetime, timezone, date, time
from decimal import Decimal
from logging import Logger
from typing import Any
from zoneinfo import ZoneInfo

from compressedfhir.utilities.json_serializers.type_preservation_serializer import (
    TypePreservationSerializer,
)


class TestCustomObject:
    def __init__(self, name: str, value: int):
        self.name: str = name
        self.value: int = value

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TestCustomObject):
            return False
        return self.name == other.name and self.value == other.value


def test_complex_data_serialization() -> None:
    """
    Test serialization and deserialization of complex data
    """
    complex_data = {
        "timestamp_no_tz": datetime.now(),
        "timestamp": datetime.now(timezone.utc),
        "timestamp_pst": datetime.now(ZoneInfo("Pacific/Honolulu")),
        "today": date.today(),
        "my_time": time(14, 30, 15),
        "my_time_pst": time(14, 30, 15, tzinfo=ZoneInfo("Pacific/Honolulu")),
        "precise_value": Decimal("3.14159"),
        "complex_number": 3 + 4j,
        "byte_data": b"Hello",
        "unique_items": {1, 2, 3},
        "custom_obj": TestCustomObject("test", 42),
    }

    # Serialize
    serialized = TypePreservationSerializer.serialize(complex_data)

    # Deserialize
    deserialized = TypePreservationSerializer.deserialize(serialized)

    # Verify types
    assert isinstance(deserialized, OrderedDict)
    assert isinstance(deserialized["timestamp_no_tz"], datetime)
    assert deserialized["timestamp_no_tz"] == complex_data["timestamp_no_tz"]
    assert isinstance(deserialized["timestamp"], datetime)
    assert deserialized["timestamp"] == complex_data["timestamp"]
    assert isinstance(deserialized["timestamp_pst"], datetime)
    assert deserialized["timestamp_pst"] == complex_data["timestamp_pst"]
    assert isinstance(deserialized["today"], date)
    assert deserialized["today"] == complex_data["today"]
    assert isinstance(deserialized["my_time"], time)
    assert deserialized["my_time"] == complex_data["my_time"]
    assert isinstance(deserialized["my_time_pst"], time)
    assert deserialized["my_time_pst"] == complex_data["my_time_pst"]
    assert isinstance(deserialized["precise_value"], Decimal)
    assert deserialized["precise_value"] == complex_data["precise_value"]
    assert isinstance(deserialized["complex_number"], complex)
    assert deserialized["complex_number"] == complex_data["complex_number"]
    assert isinstance(deserialized["byte_data"], bytes)
    assert deserialized["byte_data"] == complex_data["byte_data"]
    assert isinstance(deserialized["unique_items"], set)
    assert deserialized["unique_items"] == complex_data["unique_items"]
    assert isinstance(deserialized["custom_obj"], TestCustomObject)
    assert deserialized["custom_obj"] == complex_data["custom_obj"]


def test_nested_complex_data() -> None:
    """
    Test serialization of nested complex data
    """
    nested_data = {"level1": {"level2": {"timestamp": datetime.now(timezone.utc)}}}

    serialized = TypePreservationSerializer.serialize(nested_data)
    deserialized = TypePreservationSerializer.deserialize(serialized)

    assert isinstance(deserialized["level1"]["level2"]["timestamp"], datetime)


def test_nested_dict() -> None:
    """
    Test serialization of nested dictionaries
    """
    logger: Logger = logging.getLogger(__name__)
    nested_dict = {
        "resourceType": "Coverage",
        "id": "3456789012345670304",
        "beneficiary": {"reference": "Patient/1234567890123456703", "type": "Patient"},
        "class": [
            {
                "name": "Aetna Plan",
                "type": {
                    "coding": [
                        {
                            "code": "plan",
                            "display": "Plan",
                            "system": "http://terminology.hl7.org/CodeSystem/coverage-class",
                        }
                    ]
                },
                "value": "AE303",
            }
        ],
        "costToBeneficiary": [
            {
                "type": {"text": "Annual Physical Exams NMC - In Network"},
                "valueQuantity": {
                    "system": "http://aetna.com/Medicare/CostToBeneficiary/ValueQuantity/code",
                    "unit": "$",
                    "value": 50.0,
                },
            }
        ],
        "identifier": [
            {
                "system": "https://sources.aetna.com/coverage/identifier/membershipid/59",
                "type": {
                    "coding": [
                        {
                            "code": "SN",
                            "system": "http://terminology.hl7.org/CodeSystem/v2-0203",
                        }
                    ]
                },
                "value": "435679010300+AE303+2021-01-01",
            },
            {
                "id": "uuid",
                "system": "https://www.icanbwell.com/uuid",
                "value": "92266603-aa8b-58c6-99bd-326fd1da1896",
            },
        ],
        "meta": {
            "security": [
                {"code": "aetna", "system": "https://www.icanbwell.com/owner"},
                {"code": "aetna", "system": "https://www.icanbwell.com/access"},
                {"code": "aetna", "system": "https://www.icanbwell.com/vendor"},
                {"code": "proa", "system": "https://www.icanbwell.com/connectionType"},
            ],
            "source": "http://mock-server:1080/test_patient_access_transformer/source/4_0_0/Coverage/3456789012345670304",
        },
        "network": "Medicare - MA/NY/NJ - Full Reciprocity",
        "payor": [
            {
                "display": "Aetna",
                "reference": "Organization/6667778889990000015",
                "type": "Organization",
            }
        ],
        "period": {
            "end": datetime.fromisoformat("2021-12-31").date(),
            "start": datetime.fromisoformat("2021-01-01").date(),
        },
        "policyHolder": {"reference": "Patient/1234567890123456703", "type": "Patient"},
        "relationship": {
            "coding": [
                {
                    "code": "self",
                    "system": "http://terminology.hl7.org/CodeSystem/subscriber-relationship",
                }
            ]
        },
        "status": "active",
        "subscriber": {"reference": "Patient/1234567890123456703", "type": "Patient"},
        "subscriberId": "435679010300",
        "type": {
            "coding": [
                {
                    "code": "PPO",
                    "display": "preferred provider organization policy",
                    "system": "http://terminology.hl7.org/CodeSystem/v3-ActCode",
                }
            ]
        },
    }

    logger.info("-------- Serialized --------")
    serialized: str = TypePreservationSerializer.serialize(nested_dict)
    logger.info(serialized)
    logger.info("-------- Deserialized --------")
    deserialized: OrderedDict[str, Any] = TypePreservationSerializer.deserialize(
        serialized
    )
    logger.info(deserialized)

    assert isinstance(deserialized, OrderedDict)

    assert isinstance(deserialized["period"]["start"], date)
    assert isinstance(deserialized["period"]["end"], date)
    assert nested_dict == deserialized



import logging
from collections import OrderedDict
from datetime import datetime, date, time
from decimal import Decimal
from logging import Logger
from typing import Any, Dict, Callable, Optional, Union, cast, List
from zoneinfo import ZoneInfo


class TypePreservationDecoder:
    """
    Advanced JSON decoder for complex type reconstruction with nested type support
    """

    @classmethod
    def decode(
        cls,
        dct: Union[str, Dict[str, Any], List[Any]],
        custom_decoders: Optional[Dict[str, Callable[[Any], Any]]] = None,
        use_ordered_dict: bool = True,
    ) -> Any:
        """
        Decode complex types, including nested datetime fields

        Args:
            dct: Dictionary to decode
            custom_decoders: Optional additional custom decoders
            use_ordered_dict: Flag to control whether to use OrderedDict or not

        Returns:
            Reconstructed object or original dictionary
        """
        logger: Logger = logging.getLogger(__name__)

        # Default decoders for built-in types with nested support
        def datetime_decoder(d: Union[str, Dict[str, Any]]) -> datetime:
            if isinstance(d, str):
                return datetime.fromisoformat(d)
            elif isinstance(d, dict) and "iso" in d:
                return datetime.fromisoformat(d["iso"])
            return cast(datetime, d)

        def date_decoder(d: Union[str, Dict[str, Any]]) -> date:
            if isinstance(d, str):
                return date.fromisoformat(d)
            elif isinstance(d, dict) and "iso" in d:
                return date.fromisoformat(d["iso"])
            return cast(date, d)

        def time_decoder(d: Union[str, Dict[str, Any]]) -> time:
            if isinstance(d, str):
                return time.fromisoformat(d)
            elif isinstance(d, dict) and "iso" in d:
                # Extract ISO time string
                iso_time: str = d["iso"]

                # Parse time from ISO format
                parsed_time = time.fromisoformat(iso_time)

                # Add timezone if specified
                tz_info = d.get("tzinfo")
                if tz_info:
                    try:
                        tz_aware_time = parsed_time.replace(tzinfo=ZoneInfo(tz_info))
                        return tz_aware_time
                    except Exception as e:
                        raise ValueError(f"Invalid timezone: {tz_info}") from e
                else:
                    # If no timezone info, return naive time
                    return parsed_time
            return cast(time, d)

        default_decoders: Dict[str, Callable[[Any], Any]] = {
            "datetime": datetime_decoder,
            "date": date_decoder,
            "time": time_decoder,
            "decimal": lambda d: Decimal(d["value"] if isinstance(d, dict) else d),
            "complex": lambda d: complex(d["real"], d["imag"])
            if isinstance(d, dict)
            else d,
            "bytes": lambda d: d["value"].encode("latin-1")
            if isinstance(d, dict)
            else d,
            "set": lambda d: set(d["values"]) if isinstance(d, dict) else d,
        }

        # Merge custom decoders with default decoders
        decoders = {**default_decoders, **(custom_decoders or {})}

        # Recursively decode nested structures
        def recursive_decode(value: Any) -> Any:
            if isinstance(value, dict):
                # Check for type marker in the dictionary
                if "__type__" in value:
                    type_name = value["__type__"]

                    # Handle built-in type decoders
                    if type_name in decoders:
                        return decoders[type_name](value)

                    # Handle custom object reconstruction
                    if "__module__" in value and "attributes" in value:
                        try:
                            # Dynamically import the class
                            module = __import__(
                                value["__module__"], fromlist=[type_name]
                            )
                            cls_ = getattr(module, type_name)

                            # Create instance and set attributes with recursive decoding
                            obj = cls_.__new__(cls_)
                            obj.__dict__.update(
                                {
                                    k: recursive_decode(v)
                                    for k, v in value["attributes"].items()
                                }
                            )
                            return obj
                        except (ImportError, AttributeError) as e:
                            logger.error(f"Could not reconstruct {type_name}: {e}")
                            return value

                # Recursively decode dictionary values
                # Conditionally use OrderedDict or regular dict
                dict_type = OrderedDict if use_ordered_dict else dict
                return dict_type((k, recursive_decode(v)) for k, v in value.items())

            # Recursively decode list or tuple
            elif isinstance(value, (list, tuple)):
                return type(value)(recursive_decode(item) for item in value)

            return value

        # Start recursive decoding
        return recursive_decode(dct)



import json
from typing import Dict, Any

from compressedfhir.fhir.fhir_resource import FhirResource
from compressedfhir.utilities.compressed_dict.v1.compressed_dict_storage_mode import (
    CompressedDictStorageMode,
)


class TestFhirResource:
    def test_init_empty(self) -> None:
        """Test initializing FhirResource with no initial dictionary."""
        resource = FhirResource(storage_mode=CompressedDictStorageMode())
        assert len(resource) == 0
        assert resource.resource_type is None
        assert resource.id is None
        assert resource.resource_type_and_id is None

    def test_init_with_data(self) -> None:
        """Test initializing FhirResource with a dictionary."""
        initial_data: Dict[str, Any] = {
            "resourceType": "Patient",
            "id": "123",
            "name": [{"given": ["John"]}],
        }
        resource = FhirResource(
            initial_dict=initial_data, storage_mode=CompressedDictStorageMode()
        )

        with resource.transaction():
            assert resource.resource_type == "Patient"
            assert resource.id == "123"
            assert resource.resource_type_and_id == "Patient/123"
            assert resource["name"][0]["given"][0] == "John"

    def test_resource_type_and_id_property(self) -> None:
        """Test resource_type_and_id property with various scenarios."""
        # Scenario 1: Both resource type and id present
        resource1 = FhirResource(
            initial_dict={"resourceType": "Observation", "id": "456"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource1.resource_type_and_id == "Observation/456"

        # Scenario 2: Missing resource type
        resource2 = FhirResource(
            initial_dict={"id": "789"}, storage_mode=CompressedDictStorageMode()
        )
        assert resource2.resource_type_and_id is None

        # Scenario 3: Missing id
        resource3 = FhirResource(
            initial_dict={"resourceType": "Patient"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource3.resource_type_and_id is None

    def test_equality(self) -> None:
        """Test equality comparison between FhirResource instances."""
        # Scenario 1: Equal resources
        resource1 = FhirResource(
            initial_dict={"resourceType": "Patient", "id": "123"},
            storage_mode=CompressedDictStorageMode(),
        )
        resource2 = FhirResource(
            initial_dict={"resourceType": "Patient", "id": "123"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource1 == resource2

        # Scenario 2: Different resource types
        resource3 = FhirResource(
            initial_dict={"resourceType": "Observation", "id": "123"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource1 != resource3

        # Scenario 3: Different ids
        resource4 = FhirResource(
            initial_dict={"resourceType": "Patient", "id": "456"},
            storage_mode=CompressedDictStorageMode(),
        )
        assert resource1 != resource4

        # Scenario 4: Comparing with non-FhirResource
        assert resource1 != "Not a FhirResource"

    def test_to_json(self) -> None:
        """Test JSON serialization of FhirResource."""
        initial_data: Dict[str, Any] = {
            "resourceType": "Patient",
            "id": "123",
            "name": [{"given": ["John"]}],
        }
        resource = FhirResource(
            initial_dict=initial_data, storage_mode=CompressedDictStorageMode()
        )

        json_str = resource.json()
        parsed_json = json.loads(json_str)

        assert parsed_json == initial_data
        assert "resourceType" in parsed_json
        assert "id" in parsed_json
